import pytest

from briefcase.commands import DevCommand
from briefcase.commands.base import full_options
from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError


class DummyDevCommand(DevCommand):
    """A dummy Dev command that doesn't actually do anything.

    It only serves to track which actions would be performed.
    """

    # Platform and format contain upper case to test case normalization
    platform = "Tester"
    output_format = "Dummy"
    description = "Dummy dev command"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(*args, apps={}, **kwargs)

        self.actions = []
        self.env = dict(a=1, b=2, c=3)

    def verify_host(self):
        super().verify_host()
        self.actions.append(("verify-host",))

    def verify_tools(self):
        super().verify_tools()
        self.actions.append(("verify-tools",))

    def verify_app_template(self, app):
        super().verify_app_template(app=app)
        self.actions.append(("verify-app-template", app.app_name))

    def verify_app_tools(self, app):
        super().verify_app_tools(app=app)
        self.actions.append(("verify-app-tools", app.app_name))

    def install_dev_requirements(self, app, **kwargs):
        self.actions.append(("dev_requirements", app.app_name, kwargs))

    def get_environment(self, app, test_mode):
        return self.env

    def run_dev_app(self, app, env, **kwargs):
        self.actions.append(("run_dev", app.app_name, kwargs, env))
        return full_options({"run_dev_state": app.app_name, "env": env}, kwargs)


@pytest.fixture
def dev_command(tmp_path):
    return DummyDevCommand(base_path=tmp_path)


def test_no_args_one_app(dev_command, first_app):
    """If there is one app, dev starts that app by default."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Run the first app devly
        ("run_dev", "first", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_no_args_two_apps(dev_command, first_app, second_app):
    """If there are two apps and no explicit app provided, an error is raised."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Invoking the run command raises an error
    with pytest.raises(BriefcaseCommandError):
        dev_command(**options)

    # Finalization will not occur.
    assert dev_command.actions == []


def test_with_arg_one_app(dev_command, first_app):
    """If there is one app, and a -a argument, dev starts that app."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure a -a command line option
    options, _ = dev_command.parse_options(["-a", "first"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Run the first app devly
        ("run_dev", "first", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_with_arg_two_apps(dev_command, first_app, second_app):
    """If there are multiple apps, the --app argument starts app nominated."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure a --app command line option
    options, _ = dev_command.parse_options(["--app", "second"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified for app
        ("verify-app-tools", "second"),
        # Run the second app devly
        ("run_dev", "second", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_bad_app_reference(dev_command, first_app, second_app):
    """If the command line argument refers to an app that doesn't exist, raise an
    error."""
    # Add two apps
    dev_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure a --app command line option
    options, _ = dev_command.parse_options(["--app", "does-not-exist"])

    # Invoking the run command raises an error
    with pytest.raises(BriefcaseCommandError):
        dev_command(**options)

    # Finalization will not occur.
    assert dev_command.actions == []


def test_update_requirements(dev_command, first_app):
    """The dev command can request that the app is updated first."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure a requirements update
    options, _ = dev_command.parse_options(["-r"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # An update was requested
        ("dev_requirements", "first", {}),
        # Then, it will be started
        ("run_dev", "first", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_run_uninstalled(dev_command, first_app_uninstalled):
    """The dev command will install first if the app hasn't been installed."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure no command line options
    options, _ = dev_command.parse_options([])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # The app will be installed
        ("dev_requirements", "first", {}),
        # Then, it will be started
        ("run_dev", "first", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_update_uninstalled(dev_command, first_app_uninstalled):
    """A request to update requirements is redundant if the app hasn't been
    installed."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure a requirements update
    options, _ = dev_command.parse_options(["-r"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # An update was requested
        ("dev_requirements", "first", {}),
        # Then, it will be started
        ("run_dev", "first", {"test_mode": False, "passthrough": []}, dev_command.env),
    ]


def test_no_run(dev_command, first_app_uninstalled):
    """Install requirements without running the app."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure an update without run
    options, _ = dev_command.parse_options(["--no-run"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Only update requirements without running the app
        ("dev_requirements", "first", {}),
    ]


def test_run_test(dev_command, first_app):
    """The test suite can be run in development mode."""
    # Add a single app
    dev_command.apps = {
        "first": first_app,
    }

    # Configure the test option
    options, _ = dev_command.parse_options(["--test"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Then, it will be started
        ("run_dev", "first", {"test_mode": True, "passthrough": []}, dev_command.env),
    ]


def test_run_test_uninstalled(dev_command, first_app_uninstalled):
    """The test suite can be run in development mode."""
    # Add a single app
    dev_command.apps = {
        "first": first_app_uninstalled,
    }

    # Configure the test option
    options, _ = dev_command.parse_options(["--test"])

    # Run the run command
    dev_command(**options)

    # The right sequence of things will be done
    assert dev_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Development requirements will be installed
        ("dev_requirements", "first", {}),
        # Then, it will be started
        ("run_dev", "first", {"test_mode": True, "passthrough": []}, dev_command.env),
    ]



from __future__ import annotations

from briefcase.config import AppConfig

from .base import full_options
from .create import CreateCommand


class UpdateCommand(CreateCommand):
    command = "update"
    description = "Update the source, dependencies, and resources for an app."

    def add_options(self, parser):
        self._add_update_options(parser, update=False)
        self._add_test_options(parser, context_label="Update")

    def update_app(
        self,
        app: AppConfig,
        update_requirements: bool,
        update_resources: bool,
        update_support: bool,
        update_stub: bool,
        test_mode: bool,
        **options,
    ) -> dict | None:
        """Update an existing application bundle.

        :param app: The config object for the app
        :param update_requirements: Should requirements be updated?
        :param update_resources: Should extra resources be updated?
        :param update_support: Should app support be updated?
        :param update_stub: Should stub binary be updated?
        :param test_mode: Should the app be updated in test mode?
        """

        if not self.bundle_path(app).exists():
            self.logger.error(
                "Application does not exist; call create first!", prefix=app.app_name
            )
            return

        self.verify_app(app)

        self.logger.info("Updating application code...", prefix=app.app_name)
        self.install_app_code(app=app, test_mode=test_mode)

        if update_requirements:
            self.logger.info("Updating requirements...", prefix=app.app_name)
            self.install_app_requirements(app=app, test_mode=test_mode)

        if update_resources:
            self.logger.info("Updating application resources...", prefix=app.app_name)
            self.install_app_resources(app=app)

        if update_support:
            self.logger.info("Updating application support...", prefix=app.app_name)
            self.cleanup_app_support_package(app=app)
            self.install_app_support_package(app=app)

        if update_stub:
            try:
                # If the platform uses a stub binary, the template will define a binary
                # revision. If this template configuration item doesn't exist, there's
                # no stub binary
                self.stub_binary_revision(app)
            except KeyError:
                pass
            else:
                self.logger.info("Updating stub binary...", prefix=app.app_name)
                self.cleanup_stub_binary(app=app)
                self.install_stub_binary(app=app)

        self.logger.info("Removing unneeded app content...", prefix=app.app_name)
        self.cleanup_app_content(app=app)

        self.logger.info("Application updated.", prefix=app.app_name)

    def __call__(
        self,
        app: AppConfig | None = None,
        update_requirements: bool = False,
        update_resources: bool = False,
        update_support: bool = False,
        update_stub: bool = False,
        test_mode: bool = False,
        **options,
    ) -> dict | None:
        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        if app:
            state = self.update_app(
                app,
                update_requirements=update_requirements,
                update_resources=update_resources,
                update_support=update_support,
                update_stub=update_stub,
                test_mode=test_mode,
                **options,
            )
        else:
            state = None
            for app_name, app in sorted(self.apps.items()):
                state = self.update_app(
                    app,
                    update_requirements=update_requirements,
                    update_resources=update_resources,
                    update_support=update_support,
                    update_stub=update_stub,
                    test_mode=test_mode,
                    **full_options(state, options),
                )

        return state



import datetime
import os
import platform
import sys
import time
from pathlib import Path
from unittest import mock

import pytest
import requests

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.android_sdk import ADB, AndroidSDK
from briefcase.integrations.java import JDK
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.android.gradle import (
    GradleRunCommand,
    android_log_clean_filter,
)

from ....utils import create_file


@pytest.fixture
def jdk():
    jdk = mock.MagicMock()
    jdk.java_home = Path("/path/to/java")
    return jdk


@pytest.fixture
def run_command(tmp_path, first_app_config, jdk):
    command = GradleRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.mock_adb = mock.MagicMock(spec_set=ADB)
    command.tools.mock_adb.pidof = mock.MagicMock(return_value="777")
    command.tools.java = mock.MagicMock(spec=JDK)
    command.tools.java.java_home = "/path/to/java"
    command.tools.android_sdk = AndroidSDK(
        command.tools,
        root_path=Path("/path/to/android_sdk"),
    )
    command.tools.android_sdk.adb = mock.MagicMock(return_value=command.tools.mock_adb)

    command.tools.os = mock.MagicMock(spec_set=os)
    command.tools.os.environ = {}
    command.tools.requests = mock.MagicMock(spec_set=requests)
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.sys = mock.MagicMock(spec_set=sys)

    command._stream_app_logs = mock.MagicMock()

    command.base_path.mkdir(parents=True)
    return command


def test_binary_path(run_command, first_app_config, tmp_path):
    assert (
        run_command.binary_path(first_app_config)
        == tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "android"
        / "gradle"
        / "app"
        / "build"
        / "outputs"
        / "apk"
        / "debug"
        / "app-debug.apk"
    )


def test_device_option(run_command):
    """The -d option can be parsed."""
    options, overrides = run_command.parse_options(["-d", "myphone"])

    assert options == {
        "device_or_avd": "myphone",
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "extra_emulator_args": None,
        "shutdown_on_exit": False,
    }
    assert overrides == {}


def test_extra_emulator_args_option(run_command):
    """The -d option can be parsed."""
    options, overrides = run_command.parse_options(
        ["--Xemulator=-no-window", "--Xemulator=-no-audio"]
    )

    assert options == {
        "device_or_avd": None,
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "extra_emulator_args": ["-no-window", "-no-audio"],
        "shutdown_on_exit": False,
    }
    assert overrides == {}


def test_shutdown_on_exit_option(run_command):
    """The -d option can be parsed."""
    options, overrides = run_command.parse_options(["--shutdown-on-exit"])

    assert options == {
        "device_or_avd": None,
        "appname": None,
        "update": False,
        "update_requirements": False,
        "update_resources": False,
        "update_support": False,
        "update_stub": False,
        "no_update": False,
        "test_mode": False,
        "passthrough": [],
        "extra_emulator_args": None,
        "shutdown_on_exit": True,
    }
    assert overrides == {}


def test_unsupported_template_version(run_command, first_app_generated):
    """Error raised if template's target version is not supported."""
    run_command.apps = {"first-app": first_app_generated}

    # Skip verifying tools
    run_command.verify_tools = mock.MagicMock()

    # Mock the build command previously called
    create_file(run_command.binary_path(first_app_generated), content="")

    run_command.verify_app = mock.MagicMock(wraps=run_command.verify_app)

    run_command._briefcase_toml.update(
        {first_app_generated: {"briefcase": {"target_epoch": "0.3.16"}}}
    )

    with pytest.raises(
        BriefcaseCommandError,
        match="The app template used to generate this app is not compatible",
    ):
        run_command(first_app_generated.app_name)

    run_command.verify_app.assert_called_once_with(first_app_generated)


def test_run_existing_device(run_command, first_app_config):
    """An app can be run on an existing device."""
    # Set up device selection to return a running physical device.
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # To satisfy coverage, the stop function must be invoked at least once
    # when invoking stream_output.
    def mock_stream_output(app, stop_func, **kwargs):
        stop_func()

    run_command._stream_app_logs.side_effect = mock_stream_output

    # Set up app config to have a `-` in the `bundle`, to ensure it gets
    # normalized into a `_` via `package_name`.
    first_app_config.bundle = "com.ex-ample"

    # Invoke run_app
    run_command.run_app(
        first_app_config,
        device_or_avd="exampleDevice",
        test_mode=False,
        passthrough=[],
    )

    # select_target_device was invoked with a specific device
    run_command.tools.android_sdk.select_target_device.assert_called_once_with(
        "exampleDevice"
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="exampleDevice")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        [],
    )

    run_command.tools.mock_adb.pidof.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        quiet=True,
    )
    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The emulator was not killed at the end of the test
    run_command.tools.mock_adb.kill.assert_not_called()


def test_run_with_passthrough(run_command, first_app_config):
    """An app can be run with passthrough args."""
    # Set up device selection to return a running physical device.
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # To satisfy coverage, the stop function must be invoked at least once
    # when invoking stream_output.
    def mock_stream_output(app, stop_func, **kwargs):
        stop_func()

    run_command._stream_app_logs.side_effect = mock_stream_output

    # Set up app config to have a `-` in the `bundle`, to ensure it gets
    # normalized into a `_` via `package_name`.
    first_app_config.bundle = "com.ex-ample"

    # Invoke run_app with args.
    run_command.run_app(
        first_app_config,
        device_or_avd="exampleDevice",
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # select_target_device was invoked with a specific device
    run_command.tools.android_sdk.select_target_device.assert_called_once_with(
        "exampleDevice"
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="exampleDevice")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        ["foo", "--bar"],
    )

    run_command.tools.mock_adb.pidof.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        quiet=True,
    )
    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The emulator was not killed at the end of the test
    run_command.tools.mock_adb.kill.assert_not_called()


def test_run_slow_start(run_command, first_app_config, monkeypatch):
    """If the app is slow to start, multiple calls to pidof will be made."""
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # Mock the pidof call taking 3 attempts to return
    run_command.tools.mock_adb.pidof.side_effect = [None, None, "888"]
    monkeypatch.setattr(time, "sleep", mock.MagicMock())

    run_command.run_app(
        first_app_config,
        device_or_avd="exampleDevice",
        test_mode=False,
        passthrough=[],
    )

    assert (
        run_command.tools.mock_adb.pidof.mock_calls
        == [mock.call("com.example.first_app", quiet=True)] * 3
    )
    assert time.sleep.mock_calls == [mock.call(0.01)] * 2
    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="888")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )


def test_run_crash_at_start(run_command, first_app_config, monkeypatch):
    """If the app crashes before a PID can be read, a log dump is shown."""
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Mock the pidof call failing multiple times before a timeout.
    run_command.tools.mock_adb.pidof.side_effect = [None] * 5
    monkeypatch.setattr(time, "sleep", mock.MagicMock())

    # It's the eternal september...
    device_start_datetime = datetime.datetime(2022, 9, 9, 7, 6, 42)
    pid_lookup_start_datetime = datetime.datetime(2022, 9, 8, 7, 6, 42)

    # Mock time progression so PID lookup loops 5 times
    mock_datetime = mock.MagicMock(spec=datetime.datetime)
    # Mock datetime.now() at +0s, +0.5s, +1.5s,... +5.5s
    mock_datetime.now.side_effect = [
        pid_lookup_start_datetime + datetime.timedelta(seconds=delay)
        for delay in [0] + [x + 0.5 for x in range(0, 6)]
    ]
    monkeypatch.setattr(datetime, "datetime", mock_datetime)

    # Mock the device start datetime
    run_command.tools.mock_adb.datetime.return_value = device_start_datetime

    with pytest.raises(
        BriefcaseCommandError, match=r"Problem starting app 'first-app'"
    ):
        run_command.run_app(
            first_app_config,
            device_or_avd="exampleDevice",
            test_mode=False,
            passthrough=[],
        )

    assert (
        run_command.tools.mock_adb.pidof.mock_calls
        == [mock.call("com.example.first_app", quiet=True)] * 5
    )
    assert time.sleep.mock_calls == [mock.call(0.01)] * 5

    # The PID was never found, so logs can't be streamed
    run_command.tools.mock_adb.logcat.assert_not_called()

    # But we will get a log dump from logcat_tail
    run_command.tools.mock_adb.logcat_tail.assert_called_once_with(
        since=device_start_datetime
    )


def test_run_created_emulator(run_command, first_app_config):
    """The user can choose to run on a newly created emulator."""
    # Set up device selection to return a completely new emulator
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=(None, None, None)
    )
    run_command.tools.android_sdk.create_emulator = mock.MagicMock(
        return_value="newDevice"
    )
    run_command.tools.android_sdk.verify_avd = mock.MagicMock()
    run_command.tools.android_sdk.start_emulator = mock.MagicMock(
        return_value=("emulator-3742", "New Device")
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # Invoke run_app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # A new emulator was created
    run_command.tools.android_sdk.create_emulator.assert_called_once_with()

    # No attempt was made to verify the AVD (it is pre-verified through
    # the creation process)
    run_command.tools.android_sdk.verify_avd.assert_not_called()

    # The emulator was started
    run_command.tools.android_sdk.start_emulator.assert_called_once_with(
        "newDevice", None
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="emulator-3742")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        [],
    )

    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )


def test_run_idle_device(run_command, first_app_config):
    """The user can choose to run on an idle device."""
    # Set up device selection to return a new device that has an AVD,
    # but not a device ID.
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=(None, "Idle Device", "idleDevice")
    )

    run_command.tools.android_sdk.create_emulator = mock.MagicMock()
    run_command.tools.android_sdk.verify_avd = mock.MagicMock()
    run_command.tools.android_sdk.start_emulator = mock.MagicMock(
        return_value=("emulator-3742", "Idle Device")
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # Invoke run_app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # No attempt was made to create a new emulator
    run_command.tools.android_sdk.create_emulator.assert_not_called()

    # The AVD has been verified
    run_command.tools.android_sdk.verify_avd.assert_called_with("idleDevice")

    # The emulator was started
    run_command.tools.android_sdk.start_emulator.assert_called_once_with(
        "idleDevice", None
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="emulator-3742")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        [],
    )

    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )


def test_log_file_extra(run_command, monkeypatch):
    """Android commands register a log file extra to list SDK packages."""
    mock_android_sdk_verify = mock.MagicMock(return_value=run_command.tools.android_sdk)
    monkeypatch.setattr(AndroidSDK, "verify", mock_android_sdk_verify)
    monkeypatch.setattr(AndroidSDK, "verify_emulator", mock.MagicMock())

    # Even if one command triggers another, the sdkmanager should only be run once.
    run_command.update_command.verify_tools()
    run_command.verify_tools()

    # Android SDK tool was verified
    mock_android_sdk_verify.assert_has_calls([mock.call(tools=run_command.tools)] * 2)
    assert isinstance(run_command.tools.android_sdk, AndroidSDK)

    # list_packages() was not called
    run_command.tools.subprocess.check_output.assert_not_called()

    # list_packages() is called when saving the log
    run_command.tools.logger.save_log = True
    run_command.tools.logger.save_log_to_file(run_command)

    sdk_manager = Path(
        f"/path/to/android_sdk/cmdline-tools/{AndroidSDK.SDK_MANAGER_VER}"
        f"/bin/sdkmanager{'.bat' if platform.system() == 'Windows' else ''}"
    )
    run_command.tools.subprocess.check_output.assert_called_once_with(
        [sdk_manager, "--list_installed"],
        env={
            "ANDROID_HOME": str(run_command.tools.android_sdk.root_path),
            "ANDROID_SDK_ROOT": str(run_command.tools.android_sdk.root_path),
            "JAVA_HOME": str(run_command.tools.java.java_home),
        },
    )


def test_run_test_mode(run_command, first_app_config):
    """An app can be run in test mode."""
    # Set up device selection to return a running physical device.
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # To satisfy coverage, the stop function must be invoked at least once
    # when invoking stream_output.
    def mock_stream_output(app, stop_func, **kwargs):
        stop_func()

    run_command._stream_app_logs.side_effect = mock_stream_output

    # Set up app config to have a `-` in the `bundle`, to ensure it gets
    # normalized into a `_` via `package_name`.
    first_app_config.bundle = "com.ex-ample"

    # Invoke run_app
    run_command.run_app(
        first_app_config,
        device_or_avd="exampleDevice",
        test_mode=True,
        passthrough=[],
        shutdown_on_exit=True,
    )

    # select_target_device was invoked with a specific device
    run_command.tools.android_sdk.select_target_device.assert_called_once_with(
        "exampleDevice"
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="exampleDevice")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        [],
    )

    run_command.tools.mock_adb.pidof.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        quiet=True,
    )
    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The emulator was killed at the end of the test
    run_command.tools.mock_adb.kill.assert_called_once_with()


def test_run_test_mode_with_passthrough(run_command, first_app_config):
    """An app can be run in test mode with passthrough args."""
    # Set up device selection to return a running physical device.
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=("exampleDevice", "ExampleDevice", None)
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # To satisfy coverage, the stop function must be invoked at least once
    # when invoking stream_output.
    def mock_stream_output(app, stop_func, **kwargs):
        stop_func()

    run_command._stream_app_logs.side_effect = mock_stream_output

    # Set up app config to have a `-` in the `bundle`, to ensure it gets
    # normalized into a `_` via `package_name`.
    first_app_config.bundle = "com.ex-ample"

    # Invoke run_app in test mode with args.
    run_command.run_app(
        first_app_config,
        device_or_avd="exampleDevice",
        test_mode=True,
        passthrough=["foo", "--bar"],
        shutdown_on_exit=True,
    )

    # select_target_device was invoked with a specific device
    run_command.tools.android_sdk.select_target_device.assert_called_once_with(
        "exampleDevice"
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="exampleDevice")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        ["foo", "--bar"],
    )

    run_command.tools.mock_adb.pidof.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        quiet=True,
    )
    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The emulator was killed at the end of the test
    run_command.tools.mock_adb.kill.assert_called_once_with()


def test_run_test_mode_created_emulator(run_command, first_app_config):
    """The user can choose to run in test mode on a newly created emulator."""
    # Set up device selection to return a completely new emulator
    run_command.tools.android_sdk.select_target_device = mock.MagicMock(
        return_value=(None, None, None)
    )
    run_command.tools.android_sdk.create_emulator = mock.MagicMock(
        return_value="newDevice"
    )
    run_command.tools.android_sdk.verify_avd = mock.MagicMock()
    run_command.tools.android_sdk.start_emulator = mock.MagicMock(
        return_value=("emulator-3742", "New Device")
    )

    # Set up the log streamer to return a known stream
    log_popen = mock.MagicMock()
    run_command.tools.mock_adb.logcat.return_value = log_popen

    # Invoke run_app
    run_command.run_app(
        first_app_config,
        test_mode=True,
        passthrough=[],
        extra_emulator_args=["-no-window", "-no-audio"],
        shutdown_on_exit=True,
    )

    # A new emulator was created
    run_command.tools.android_sdk.create_emulator.assert_called_once_with()

    # No attempt was made to verify the AVD (it is pre-verified through
    # the creation process)
    run_command.tools.android_sdk.verify_avd.assert_not_called()

    # The emulator was started
    run_command.tools.android_sdk.start_emulator.assert_called_once_with(
        "newDevice",
        ["-no-window", "-no-audio"],
    )

    # The ADB wrapper is created
    run_command.tools.android_sdk.adb.assert_called_once_with(device="emulator-3742")

    # The adb wrapper is invoked with the expected arguments
    run_command.tools.mock_adb.install_apk.assert_called_once_with(
        run_command.binary_path(first_app_config)
    )
    run_command.tools.mock_adb.force_stop_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
    )

    run_command.tools.mock_adb.start_app.assert_called_once_with(
        f"{first_app_config.package_name}.{first_app_config.module_name}",
        "org.beeware.android.MainActivity",
        [],
    )

    run_command.tools.mock_adb.logcat.assert_called_once_with(pid="777")

    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_filter=android_log_clean_filter,
        clean_output=False,
        stop_func=mock.ANY,
        log_stream=True,
    )

    # The emulator was killed at the end of the test
    run_command.tools.mock_adb.kill.assert_called_once_with()



import subprocess

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_verify_repo(flatpak, tool_debug_mode):
    """A Flatpak repo can be verified."""
    # Enable verbose tool logging
    if tool_debug_mode:
        flatpak.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    flatpak.verify_repo(
        repo_alias="test-alias",
        url="https://example.com/flatpak",
    )

    # The expected call was made
    flatpak.tools.subprocess.run.assert_called_once_with(
        [
            "flatpak",
            "remote-add",
            "--user",
            "--if-not-exists",
            "test-alias",
            "https://example.com/flatpak",
        ]
        + (["--verbose"] if tool_debug_mode else []),
        check=True,
    )


def test_verify_repo_fail(flatpak):
    """If repo verification fails, an error is raised."""
    flatpak.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd="flatpak repo-add", returncode=1
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to add Flatpak repo https://example.com/flatpak with alias test-alias.",
    ):
        flatpak.verify_repo(
            repo_alias="test-alias",
            url="https://example.com/flatpak",
        )



from pathlib import Path
from subprocess import CalledProcessError
from unittest import mock
from zipfile import ZipFile

import pytest

import briefcase.platforms.windows
from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.integrations.windows_sdk import WindowsSDK
from briefcase.integrations.wix import WiX
from briefcase.platforms.windows.app import WindowsAppPackageCommand

from ....utils import create_file


@pytest.fixture
def package_command(tmp_path):
    command = WindowsAppPackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.host_os = "Windows"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    command.tools.wix = WiX(command.tools, wix_home=tmp_path / "wix")
    command.tools.windows_sdk = WindowsSDK(
        tools=command.tools,
        root_path=tmp_path / "windows_sdk",
        version="81.2.1.0",
        arch="groovy",
    )
    return command


@pytest.fixture
def package_command_with_files(package_command, tmp_path):
    # Build the paths for the source and the distribution folders:
    src_path = tmp_path / "base_path/build/first-app/windows/app/src"
    dist_path = tmp_path / "base_path/dist"
    src_path.mkdir(parents=True)
    dist_path.mkdir(parents=True)

    # Mock some typical folders and files in the src folder:
    files = (
        src_path / "First App.exe",
        src_path / "python.exe",
        src_path / "python3.dll",
        src_path / "vcruntime140.dll",
        src_path / "app/first-app/app.py",
        src_path / "app/first-app/resources/__init__.py",
        src_path / "app/first-app-0.0.1.dist-info/top_level.txt",
        src_path / "app_packages/clr.py",
        src_path / "app_packages/toga_winforms/command.py",
    )
    for file in files:
        create_file(file, "")

    return package_command


def test_package_formats(package_command):
    """Packaging formats are as expected."""
    assert package_command.packaging_formats == ["msi", "zip"]
    assert package_command.default_packaging_format == "msi"


def test_verify(package_command, monkeypatch):
    """Verifying on Windows creates a WiX wrapper."""
    # prime Command to _not_ need Windows SDK
    package_command._windows_sdk_needed = False

    mock_wix_verify = mock.MagicMock(wraps=WiX.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WiX,
        "verify",
        mock_wix_verify,
    )

    package_command.verify_tools()

    # WiX tool was verified
    mock_wix_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.wix, WiX)


def test_verify_with_signing(package_command, monkeypatch):
    """Verifying on Windows creates WiX and WindowsSDK wrappers when code signing."""
    # prime Command to need Windows SDK
    package_command._windows_sdk_needed = True

    mock_windows_sdk_verify = mock.MagicMock(wraps=WindowsSDK.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WindowsSDK,
        "verify",
        mock_windows_sdk_verify,
    )

    mock_wix_verify = mock.MagicMock(wraps=WiX.verify)
    monkeypatch.setattr(
        briefcase.platforms.windows.WiX,
        "verify",
        mock_wix_verify,
    )

    package_command.verify_tools()

    # WiX tool was verified
    mock_wix_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.wix, WiX)
    # WindowsSDK tool was verified
    mock_windows_sdk_verify.assert_called_once_with(tools=package_command.tools)
    assert isinstance(package_command.tools.windows_sdk, WindowsSDK)


@pytest.mark.parametrize(
    "cli_args, signing_options, is_sdk_needed",
    [
        ([], {}, False),
        (["--adhoc-sign"], dict(adhoc_sign=True), False),
        (["--file-digest", "sha2000"], dict(file_digest="sha2000"), False),
        (["-i", "asdf"], dict(identity="asdf"), True),
        (["--identity", "asdf"], dict(identity="asdf"), True),
        (["--identity", "asdf"], dict(identity="asdf"), True),
        (
            [
                "-i",
                "asdf",
                "--file-digest",
                "sha42",
                "--use-local-machine-stores",
                "--cert-store",
                "mystore",
                "--timestamp-url",
                "http://freetimestamps.com",
                "--timestamp-digest",
                "sha56",
            ],
            dict(
                identity="asdf",
                file_digest="sha42",
                use_local_machine=True,
                cert_store="mystore",
                timestamp_url="http://freetimestamps.com",
                timestamp_digest="sha56",
            ),
            True,
        ),
        (
            [
                "-i",
                "asdf",
                "--file-digest",
                "sha42",
                "--cert-store",
                "mystore",
                "--timestamp-url",
                "http://freetimestamps.com",
                "--timestamp-digest",
                "sha56",
            ],
            dict(
                identity="asdf",
                file_digest="sha42",
                cert_store="mystore",
                timestamp_url="http://freetimestamps.com",
                timestamp_digest="sha56",
            ),
            True,
        ),
    ],
)
def test_parse_options(package_command, cli_args, signing_options, is_sdk_needed):
    """Command line arguments are parsed as expected; Windows SDK is required if an
    identity is specified."""
    default_options = dict(
        identity=None,
        file_digest="sha256",
        use_local_machine=False,
        cert_store="My",
        timestamp_url="http://timestamp.digicert.com",
        timestamp_digest="sha256",
        adhoc_sign=False,
        packaging_format="msi",
        update=False,
    )
    expected_options = {**default_options, **signing_options}

    options, overrides = package_command.parse_options(extra=cli_args)

    assert options == expected_options
    assert overrides == {}
    assert package_command._windows_sdk_needed is is_sdk_needed


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),  # Default behavior (adhoc signing)
        {"adhoc_sign": True},  # Explicit adhoc signing
    ],
)
def test_package_msi(package_command, first_app_config, kwargs, tmp_path):
    """A Windows app can be packaged as an MSI."""

    package_command.package_app(first_app_config, **kwargs)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix/bin/candle.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-arch",
                "x64",
                "-dSourceDir=src",
                "first-app.wxs",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Link MSI
        mock.call(
            [
                tmp_path / "wix/bin/light.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-loc",
                "unicode.wxl",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
                "first-app.wixobj",
                "first-app-manifest.wixobj",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),  # Default behavior (adhoc signing)
        {"adhoc_sign": True},  # Explicit adhoc signing
    ],
)
def test_package_zip(package_command_with_files, first_app_config, kwargs, tmp_path):
    """A Windows app can be packaged as a zip file."""

    first_app_config.packaging_format = "zip"
    package_command_with_files.package_app(first_app_config, **kwargs)

    # No signing was performed
    assert package_command_with_files.tools.subprocess.run.mock_calls == []

    archive_file = tmp_path / "base_path/dist/First App-0.0.1.zip"
    source_folders_and_files = (
        "app/",
        "app/first-app/",
        "app/first-app/resources/",
        "app/first-app-0.0.1.dist-info/",
        "app_packages/",
        "app_packages/toga_winforms/",
        "First App.exe",
        "python.exe",
        "python3.dll",
        "vcruntime140.dll",
        "app/first-app/app.py",
        "app/first-app/resources/__init__.py",
        "app/first-app-0.0.1.dist-info/top_level.txt",
        "app_packages/clr.py",
        "app_packages/toga_winforms/command.py",
    )

    # The zip file exists
    assert archive_file.exists()

    # Check content of zip file
    with ZipFile(archive_file) as archive:
        root = "First App-0.0.1/"
        # All folders and files in zip are from source
        for name in archive.namelist():
            # name.removeprefix(f'{root}') will only work in Python > 3.8
            assert name[len(root) :] in source_folders_and_files
        # All files from source are in zip
        for file in source_folders_and_files:
            assert f"{root}{file}" in archive.namelist()


@pytest.mark.parametrize(
    "use_local_machine, additional_args",
    [(False, []), (True, ["-sm"])],
)
def test_package_msi_with_codesigning(
    package_command,
    first_app_config,
    tmp_path,
    use_local_machine,
    additional_args,
):
    """A Windows app can be packaged as an MSI and code signed."""

    package_command.package_app(
        first_app_config,
        identity="80ee4c3321122916f5637522451993c2a0a4a56a",
        file_digest="sha42",
        use_local_machine=use_local_machine,
        cert_store="mystore",
        timestamp_url="http://freetimestamps.com",
        timestamp_digest="sha56",
    )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
            ]
            + additional_args
            + [
                tmp_path
                / "base_path"
                / "build"
                / "first-app"
                / "windows"
                / "app"
                / "src"
                / "First App.exe"
            ],
            check=True,
        ),
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix/bin/candle.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-arch",
                "x64",
                "-dSourceDir=src",
                "first-app.wxs",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Link MSI
        mock.call(
            [
                tmp_path / "wix/bin/light.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-loc",
                "unicode.wxl",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
                "first-app.wixobj",
                "first-app-manifest.wixobj",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Codesign app MSI
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
            ]
            + additional_args
            + [tmp_path / "base_path/dist/First App-0.0.1.msi"],
            check=True,
        ),
    ]


@pytest.mark.parametrize(
    "use_local_machine, additional_args",
    [(False, []), (True, ["-sm"])],
)
def test_package_zip_with_codesigning(
    package_command_with_files,
    first_app_config,
    tmp_path,
    use_local_machine,
    additional_args,
):
    """In a ZIP package, only the binary will be code signed."""

    first_app_config.packaging_format = "zip"

    package_command_with_files.package_app(
        first_app_config,
        identity="80ee4c3321122916f5637522451993c2a0a4a56a",
        file_digest="sha42",
        use_local_machine=use_local_machine,
        cert_store="mystore",
        timestamp_url="http://freetimestamps.com",
        timestamp_digest="sha56",
    )

    assert package_command_with_files.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
            ]
            + additional_args
            + [
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        ),
    ]


def test_package_msi_invalid_identity(package_command, first_app_config):
    """Codesigning fails, along with packaging, if the identity is invalid."""
    with pytest.raises(
        BriefcaseCommandError,
        match="Codesigning identify 'asdf' must be a certificate SHA-1 thumbprint",
    ):
        package_command.package_app(first_app_config, identity="asdf")


def test_package_msi_failed_sign_app(package_command, first_app_config, tmp_path):
    """An error is raised if signtool fails for the app exe."""
    # Mock a failure in the call to signtool.exe
    package_command.tools.subprocess.run.side_effect = [
        CalledProcessError(cmd=["signtool.exe"], returncode=1),
    ]

    with pytest.raises(BriefcaseCommandError, match=r"Unable to sign "):
        package_command.package_app(
            first_app_config,
            identity="80ee4c3321122916f5637522451993c2a0a4a56a",
            file_digest="sha42",
            cert_store="mystore",
            timestamp_url="http://freetimestamps.com",
            timestamp_digest="sha56",
        )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        )
    ]


def test_package_msi_failed_manifest(package_command, first_app_config, tmp_path):
    """An error is raised if a manifest cannot be built."""
    # Mock a failure in the call to heat.exe
    package_command.tools.subprocess.run.side_effect = [
        CalledProcessError(cmd=["heat.exe"], returncode=1),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to generate manifest for app first-app.",
    ):
        package_command.package_app(first_app_config)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


def test_package_msi_failed_compile(package_command, first_app_config, tmp_path):
    """An error is raised if compilation failed."""
    # Mock a failure in the call to candle.exe
    package_command.tools.subprocess.run.side_effect = [
        None,  # heat.exe
        CalledProcessError(cmd=["candle.exe"], returncode=1),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to compile app first-app.",
    ):
        package_command.package_app(first_app_config)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix/bin/candle.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-arch",
                "x64",
                "-dSourceDir=src",
                "first-app.wxs",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


def test_package_msi_failed_link(package_command, first_app_config, tmp_path):
    """An error is raised if linking fails."""
    # Mock a failure in the call to light.exe
    package_command.tools.subprocess.run.side_effect = [
        None,  # heat.exe
        None,  # candle.exe
        CalledProcessError(cmd=["link.exe"], returncode=1),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to link app first-app.",
    ):
        package_command.package_app(first_app_config)

    assert package_command.tools.subprocess.run.mock_calls == [
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix/bin/candle.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-arch",
                "x64",
                "-dSourceDir=src",
                "first-app.wxs",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Link MSI
        mock.call(
            [
                tmp_path / "wix/bin/light.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-loc",
                "unicode.wxl",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
                "first-app.wixobj",
                "first-app-manifest.wixobj",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
    ]


def test_package_msi_failed_signing_msi(package_command, first_app_config, tmp_path):
    """An error is raised if signtool fails for the app MSI."""
    # Mock a failure in the call to light.exe
    package_command.tools.subprocess.run.side_effect = [
        None,  # signtool.exe
        None,  # heat.exe
        None,  # candle.exe
        None,  # link.exe
        CalledProcessError(cmd=["signtool.exe"], returncode=1),
    ]

    with pytest.raises(BriefcaseCommandError, match=r"Unable to sign "):
        package_command.package_app(
            first_app_config,
            identity="80ee4c3321122916f5637522451993c2a0a4a56a",
            file_digest="sha42",
            cert_store="mystore",
            timestamp_url="http://freetimestamps.com",
            timestamp_digest="sha56",
        )

    assert package_command.tools.subprocess.run.mock_calls == [
        # Codesign app exe
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/build/first-app/windows/app/src/First App.exe",
            ],
            check=True,
        ),
        # Collect manifest
        mock.call(
            [
                tmp_path / "wix/bin/heat.exe",
                "dir",
                Path("src"),
                "-nologo",
                "-gg",
                "-sfrag",
                "-sreg",
                "-srd",
                "-scom",
                "-dr",
                "first_app_ROOTDIR",
                "-cg",
                "first_app_COMPONENTS",
                "-var",
                "var.SourceDir",
                "-out",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Compile MSI
        mock.call(
            [
                tmp_path / "wix/bin/candle.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-arch",
                "x64",
                "-dSourceDir=src",
                "first-app.wxs",
                "first-app-manifest.wxs",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Link MSI
        mock.call(
            [
                tmp_path / "wix/bin/light.exe",
                "-nologo",
                "-ext",
                "WixUtilExtension",
                "-ext",
                "WixUIExtension",
                "-loc",
                "unicode.wxl",
                "-o",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
                "first-app.wixobj",
                "first-app-manifest.wixobj",
            ],
            check=True,
            cwd=tmp_path / "base_path/build/first-app/windows/app",
        ),
        # Codesign app MSI
        mock.call(
            [
                tmp_path
                / "windows_sdk"
                / "bin"
                / "81.2.1.0"
                / "groovy"
                / "signtool.exe",
                "sign",
                "-s",
                "mystore",
                "-sha1",
                "80ee4c3321122916f5637522451993c2a0a4a56a",
                "-fd",
                "sha42",
                "-d",
                "The first simple app \\ demonstration",
                "-du",
                "https://example.com/first-app",
                "-tr",
                "http://freetimestamps.com",
                "-td",
                "sha56",
                tmp_path / "base_path/dist/First App-0.0.1.msi",
            ],
            check=True,
        ),
    ]



from unittest.mock import MagicMock

import briefcase.console
from briefcase.console import NotDeadYet, Printer


def test_update(capsys, monkeypatch):
    """The message is only printed once for each interval."""
    # initialization will set interval to 0 + 10
    # update() will see time either at 5 or 15 and print the message accordingly
    # then interval is reset back to 0 + 10
    monkeypatch.setattr(
        briefcase.console.time,
        "time",
        MagicMock(side_effect=[0] + [10 - 5, 10 + 5, 0] * 3),
    )

    keep_alive = NotDeadYet(printer=Printer())

    for _ in range(3):
        keep_alive.update()
        keep_alive.update()

    assert capsys.readouterr().out == (
        "... still waiting\n... still waiting\n... still waiting\n"
    )


def test_reset(capsys, monkeypatch):
    """Calling reset always puts the interval in the future and nothing prints."""
    # initialization will set interval to 0 + 10
    # the reset updates the interval to 10 + 10
    # the update sees a time of 15 and doesn't print or reset
    monkeypatch.setattr(
        briefcase.console.time,
        "time",
        MagicMock(side_effect=[0] + [10, 15] * 10),
    )

    keep_alive = NotDeadYet(printer=Printer())

    for _ in range(10):
        keep_alive.reset()
        keep_alive.update()

    assert capsys.readouterr().out == ""



from unittest.mock import MagicMock

import pytest

import briefcase.commands.upgrade
from briefcase.commands import UpgradeCommand
from briefcase.console import Console, Log
from briefcase.exceptions import MissingToolError
from briefcase.integrations.base import ManagedTool, Tool


@pytest.fixture
def upgrade_command(tmp_path):
    command = DummyUpgradeCommand(base_path=tmp_path)
    command.tools.host_os = "wonky"
    return command


class DummyUpgradeCommand(UpgradeCommand):
    """A dummy upgrade command that doesn't actually do anything.

    It only serves to track which actions would be performed.
    """

    # Platform and format contain upper case to test case normalization
    platform = "Tester"
    output_format = "Dummy"
    description = "Dummy update command"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(*args, **kwargs)

    def bundle_path(self, app):
        return self.platform_path / f"{app.app_name}.dummy"

    def binary_path(self, app):
        return self.bundle_path(app) / f"{app.app_name}.bin"


@pytest.fixture
def mock_tool_registry(monkeypatch):
    """Tool registry with all dummy tools."""
    tool_list = [
        DummyTool,
        DummyManagedTool1,
        DummyManagedTool2,
        DummyManagedTool3,
        DummyUnManagedManagedTool,
        DummyNotInstalledManagedTool,
    ]

    tool_registry = dict()
    for tool in tool_list:
        monkeypatch.setattr(tool, "verify", MagicMock(wraps=tool.verify))
        tool_registry[tool.name] = tool

    monkeypatch.setattr(briefcase.commands.upgrade, "tool_registry", tool_registry)


@pytest.fixture
def mock_no_managed_tool_registry(monkeypatch):
    """Tool registry without any installed managed tools."""
    tool_list = [
        DummyTool,
        DummyUnManagedManagedTool,
        DummyNotInstalledManagedTool,
    ]

    tool_registry = dict()
    for tool in tool_list:
        monkeypatch.setattr(tool, "verify", MagicMock(wraps=tool.verify))
        tool_registry[tool.name] = tool

    monkeypatch.setattr(briefcase.commands.upgrade, "tool_registry", tool_registry)


class DummyToolBase(Tool):
    name = "dummy_tool_base"
    supported_host_os = {"wonky"}

    @classmethod
    def verify_install(cls, tools, **kwargs):
        return cls(tools=tools)


class DummyManagedToolBase(ManagedTool):
    name = "dummy_managed_tool_base"
    supported_host_os = {"wonky"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.actions = []

    @classmethod
    def verify_install(cls, tools, **kwargs):
        # add to ToolCache so accessible after upgrade
        setattr(tools, cls.name, cls(tools=tools))
        return getattr(tools, cls.name)

    def exists(self) -> bool:
        self.actions.append("exists")
        return True

    def install(self):
        self.actions.append("install")

    def uninstall(self):
        self.actions.append("uninstall")


class DummyTool(DummyToolBase):
    """Unmanaged Tool testing class."""

    name = "unmanaged"
    full_name = "Unmanaged Dummy Tool"


class DummyUnManagedManagedTool(DummyManagedToolBase):
    """Managed Tool without a managed install testing class."""

    name = "unmanaged_managed"
    full_name = "Unmanaged Managed Dummy Tool"

    @property
    def managed_install(self) -> bool:
        return False


class DummyNotInstalledManagedTool(DummyManagedToolBase):
    """Managed Tool without a managed install testing class."""

    name = "not_installed"
    full_name = "Not Installed Managed Dummy Tool"

    @classmethod
    def verify_install(cls, tools, **kwargs):
        raise MissingToolError(cls.full_name)


class DummyManagedTool1(DummyManagedToolBase):
    """Managed Tool testing class."""

    name = "managed_1"
    full_name = "Managed Dummy Tool 1"


class DummyManagedTool2(DummyManagedToolBase):
    """Managed Tool testing class."""

    name = "managed_2"
    full_name = "Managed Dummy Tool 2"


class DummyManagedTool3(DummyManagedToolBase):
    """Managed Tool testing class."""

    name = "managed_3"
    full_name = "Managed Dummy Tool 3"



import plistlib
from unittest import mock

import pytest

from briefcase.exceptions import BriefcaseCommandError

from .....utils import create_file


@pytest.fixture
def license_file(tmp_path):
    path = tmp_path / "base_path/LICENSE"
    create_file(path, "You can take license with this.")
    return path


def test_gui_app(
    package_command,
    first_app_with_binaries,
    license_file,
    sekrit_identity,
    sekrit_installer_identity,
    tmp_path,
):
    """A macOS GUI app can be packaged as a .pkg installer."""
    first_app_with_binaries.packaging_format = "pkg"

    # Select a codesigning identity
    package_command.select_identity.side_effect = [
        sekrit_identity,
        sekrit_installer_identity,
    ]

    # Mock the notarization process
    package_command.notarize = mock.Mock()

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Create a pre-existing app bundle.
    create_file(
        bundle_path / "installer/root/First App.app/original",
        "Original app",
    )

    # Create a pre-existing package bundle.
    create_file(
        bundle_path / "installer/packages/first-app.pkg",
        "Original package",
    )

    # Create a pre-existing LICENSE
    create_file(
        bundle_path / "installer/resources/LICENSE",
        "Original License",
    )

    # Re-package the app
    package_command.package_app(first_app_with_binaries)

    # Two signing identities were selected; the second is an installer identity
    package_command.select_identity.mock_calls == [
        mock.call(identity=None),
        mock.call(identity=None, app_identity=sekrit_identity),
    ]

    # The app has been signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=sekrit_identity,
    )

    # App content has been copied into place.
    assert (bundle_path / "installer/root/First App.app/Contents/Info.plist").is_file()

    # The license has been updated.
    assert (bundle_path / "installer/resources/LICENSE").read_text(
        encoding="utf-8"
    ) == "You can take license with this."

    # The component list has been updated.
    with (bundle_path / "installer/components.plist").open("rb") as f:
        components = plistlib.load(f)

        assert components == [
            {
                "BundleHasStrictIdentifier": True,
                "BundleIsRelocatable": False,
                "BundleIsVersionChecked": True,
                "BundleOverwriteAction": "upgrade",
                "RootRelativeBundlePath": "First App.app",
            }
        ]

    assert package_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                "pkgbuild",
                "--root",
                bundle_path / "installer/root",
                "--component-plist",
                bundle_path / "installer/components.plist",
                "--install-location",
                "/Applications",
                bundle_path / "installer/packages/first-app.pkg",
            ],
            check=True,
        ),
        mock.call(
            [
                "productbuild",
                "--distribution",
                bundle_path / "installer/Distribution.xml",
                "--package-path",
                bundle_path / "installer/packages",
                "--resources",
                bundle_path / "installer/resources",
                "--sign",
                "CAFEFACE",
                tmp_path / "base_path/dist/First App-0.0.1.pkg",
            ],
            check=True,
        ),
    ]

    # Notarization was performed with the installer identity
    package_command.notarize.assert_called_once_with(
        tmp_path / "base_path/dist/First App-0.0.1.pkg",
        identity=sekrit_installer_identity,
    )


def test_gui_app_adhoc_identity(
    package_command,
    first_app_with_binaries,
    license_file,
    adhoc_identity,
    tmp_path,
):
    """A macOS GUI app can be packaged as a .pkg installer."""
    first_app_with_binaries.packaging_format = "pkg"

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Mock the notarization process
    package_command.notarize = mock.Mock()

    # Create a pre-existing app bundle.
    create_file(
        bundle_path / "installer/root/First App.app/original",
        "Original app",
    )

    # Create a pre-existing package bundle.
    create_file(
        bundle_path / "installer/packages/first-app.pkg",
        "Original package",
    )

    # Create a pre-existing LICENSE
    create_file(
        bundle_path / "installer/resources/LICENSE",
        "Original License",
    )

    # Re-package the app
    package_command.package_app(
        first_app_with_binaries,
        adhoc_sign=True,
    )

    # The app has been signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=adhoc_identity,
    )

    # App content has been copied into place.
    assert (bundle_path / "installer/root/First App.app/Contents/Info.plist").is_file()

    # The license has been updated.
    assert (bundle_path / "installer/resources/LICENSE").read_text(
        encoding="utf-8"
    ) == "You can take license with this."

    # The component list has been updated.
    with (bundle_path / "installer/components.plist").open("rb") as f:
        components = plistlib.load(f)

        assert components == [
            {
                "BundleHasStrictIdentifier": True,
                "BundleIsRelocatable": False,
                "BundleIsVersionChecked": True,
                "BundleOverwriteAction": "upgrade",
                "RootRelativeBundlePath": "First App.app",
            }
        ]

    assert package_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                "pkgbuild",
                "--root",
                bundle_path / "installer/root",
                "--component-plist",
                bundle_path / "installer/components.plist",
                "--install-location",
                "/Applications",
                bundle_path / "installer/packages/first-app.pkg",
            ],
            check=True,
        ),
        mock.call(
            [
                "productbuild",
                "--distribution",
                bundle_path / "installer/Distribution.xml",
                "--package-path",
                bundle_path / "installer/packages",
                "--resources",
                bundle_path / "installer/resources",
                tmp_path / "base_path/dist/First App-0.0.1.pkg",
            ],
            check=True,
        ),
    ]

    # No notarization was performed
    package_command.notarize.assert_not_called()


def test_console_app(
    package_command,
    first_app_with_binaries,
    license_file,
    sekrit_identity,
    sekrit_installer_identity,
    tmp_path,
):
    """A macOS console app can be packaged as a .pkg installer."""
    first_app_with_binaries.packaging_format = "pkg"
    first_app_with_binaries.console_app = True

    # Select a codesigning identity
    package_command.select_identity.side_effect = [
        sekrit_identity,
        sekrit_installer_identity,
    ]

    # Mock the notarization process
    package_command.notarize = mock.Mock()

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Package the app
    package_command.package_app(first_app_with_binaries)

    # Two signing identities were selected; the second is an installer identity
    package_command.select_identity.mock_calls == [
        mock.call(identity=None),
        mock.call(identity=None, app_identity=sekrit_identity),
    ]

    # The app has been signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=sekrit_identity,
    )

    # App content has been copied into place.
    assert (bundle_path / "installer/root/First App.app/Contents/Info.plist").is_file()

    # The license has been installed
    assert (bundle_path / "installer/resources/LICENSE").read_text(
        encoding="utf-8"
    ) == "You can take license with this."

    # The component list has been updated.
    with (bundle_path / "installer/components.plist").open("rb") as f:
        components = plistlib.load(f)

        assert components == [
            {
                "BundleHasStrictIdentifier": True,
                "BundleIsRelocatable": False,
                "BundleIsVersionChecked": True,
                "BundleOverwriteAction": "upgrade",
                "RootRelativeBundlePath": "First App.app",
            }
        ]

    assert package_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                "pkgbuild",
                "--root",
                bundle_path / "installer/root",
                "--component-plist",
                bundle_path / "installer/components.plist",
                "--install-location",
                "/Library/First App",
                "--scripts",
                bundle_path / "installer/scripts",
                bundle_path / "installer/packages/first-app.pkg",
            ],
            check=True,
        ),
        mock.call(
            [
                "productbuild",
                "--distribution",
                bundle_path / "installer/Distribution.xml",
                "--package-path",
                bundle_path / "installer/packages",
                "--resources",
                bundle_path / "installer/resources",
                "--sign",
                "CAFEFACE",
                tmp_path / "base_path/dist/First App-0.0.1.pkg",
            ],
            check=True,
        ),
    ]

    # Notarization was performed with the installer identity
    package_command.notarize.assert_called_once_with(
        tmp_path / "base_path/dist/First App-0.0.1.pkg",
        identity=sekrit_installer_identity,
    )


def test_console_app_adhoc_signed(
    package_command,
    first_app_with_binaries,
    license_file,
    adhoc_identity,
    tmp_path,
):
    """A macOS console app can be packaged as a .pkg installer."""
    first_app_with_binaries.packaging_format = "pkg"
    first_app_with_binaries.console_app = True

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Mock the notarization process
    package_command.notarize = mock.Mock()

    # Package the app
    package_command.package_app(
        first_app_with_binaries,
        adhoc_sign=True,
    )

    # The app has been signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=adhoc_identity,
    )

    # App content has been copied into place.
    assert (bundle_path / "installer/root/First App.app/Contents/Info.plist").is_file()

    # The license has been installed
    assert (bundle_path / "installer/resources/LICENSE").read_text(
        encoding="utf-8"
    ) == "You can take license with this."

    # The component list has been updated.
    with (bundle_path / "installer/components.plist").open("rb") as f:
        components = plistlib.load(f)

        assert components == [
            {
                "BundleHasStrictIdentifier": True,
                "BundleIsRelocatable": False,
                "BundleIsVersionChecked": True,
                "BundleOverwriteAction": "upgrade",
                "RootRelativeBundlePath": "First App.app",
            }
        ]

    assert package_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                "pkgbuild",
                "--root",
                bundle_path / "installer/root",
                "--component-plist",
                bundle_path / "installer/components.plist",
                "--install-location",
                "/Library/First App",
                "--scripts",
                bundle_path / "installer/scripts",
                bundle_path / "installer/packages/first-app.pkg",
            ],
            check=True,
        ),
        mock.call(
            [
                "productbuild",
                "--distribution",
                bundle_path / "installer/Distribution.xml",
                "--package-path",
                bundle_path / "installer/packages",
                "--resources",
                bundle_path / "installer/resources",
                tmp_path / "base_path/dist/First App-0.0.1.pkg",
            ],
            check=True,
        ),
    ]

    # No notarization was performed
    package_command.notarize.assert_not_called()


def test_no_license(package_command, first_app_with_binaries, adhoc_identity, tmp_path):
    """If the project has no license file, an error is raised."""
    first_app_with_binaries.packaging_format = "pkg"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Your project does not contain a LICENSE file",
    ):
        package_command.package_app(
            first_app_with_binaries,
            adhoc_sign=True,
        )

    # The app will be signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=adhoc_identity,
    )

    # Component manifest hasn't been written
    assert not (
        tmp_path / "base_path/build/first-app/macos/app/installer/components.plist"
    ).exists()

    # No calls made to pkgbuild/productbuild
    package_command.tools.subprocess.run.assert_not_called()


def test_package_pkg_previously_built(
    package_command,
    first_app_with_binaries,
    license_file,
    adhoc_identity,
    tmp_path,
):
    """If a previous installer was built, the package folder is recreated."""
    first_app_with_binaries.packaging_format = "pkg"

    bundle_path = tmp_path / "base_path/build/first-app/macos/app"

    # Create a pre-existing app bundle.
    create_file(
        bundle_path / "installer/root/First App.app/original",
        "Original app",
    )

    # Create a pre-existing package bundle.
    create_file(
        bundle_path / "installer/packages/first-app.pkg",
        "Original package",
    )

    # Create a pre-existing LICENSE
    create_file(
        bundle_path / "installer/resources/LICENSE",
        "Original License",
    )

    # Re-package the app
    package_command.package_app(
        first_app_with_binaries,
        adhoc_sign=True,
    )

    # The app has been signed
    package_command.sign_app.assert_called_once_with(
        app=first_app_with_binaries,
        identity=adhoc_identity,
    )

    # App content has been copied into place.
    assert (bundle_path / "installer/root/First App.app/Contents/Info.plist").is_file()

    # ... but the old file doesn't exist
    assert not (bundle_path / "installer/root/First App.app/original").exists()

    # The old package data doesn't exist either
    assert not (bundle_path / "installer/packages/first-app.pkg").exists()

    # The license has been updated.
    assert (bundle_path / "installer/resources/LICENSE").read_text(
        encoding="utf-8"
    ) == "You can take license with this."

    # The component list has been updated.
    with (bundle_path / "installer/components.plist").open("rb") as f:
        components = plistlib.load(f)

        assert components == [
            {
                "BundleHasStrictIdentifier": True,
                "BundleIsRelocatable": False,
                "BundleIsVersionChecked": True,
                "BundleOverwriteAction": "upgrade",
                "RootRelativeBundlePath": "First App.app",
            }
        ]

    assert package_command.tools.subprocess.run.mock_calls == [
        mock.call(
            [
                "pkgbuild",
                "--root",
                bundle_path / "installer/root",
                "--component-plist",
                bundle_path / "installer/components.plist",
                "--install-location",
                "/Applications",
                bundle_path / "installer/packages/first-app.pkg",
            ],
            check=True,
        ),
        mock.call(
            [
                "productbuild",
                "--distribution",
                bundle_path / "installer/Distribution.xml",
                "--package-path",
                bundle_path / "installer/packages",
                "--resources",
                bundle_path / "installer/resources",
                tmp_path / "base_path/dist/First App-0.0.1.pkg",
            ],
            check=True,
        ),
    ]



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_list_packages(mock_tools, android_sdk):
    """Sdkmanager can be used to list the packages managed by the Android SDK."""
    android_sdk.list_packages()

    mock_tools.subprocess.check_output.assert_called_once_with(
        [android_sdk.sdkmanager_path, "--list_installed"],
        env=android_sdk.env,
    )


def test_list_packages_failure(mock_tools, android_sdk):
    """If sdkmanager fails, an error is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        1, ""
    )
    with pytest.raises(BriefcaseCommandError):
        android_sdk.list_packages()

    mock_tools.subprocess.check_output.assert_called_once_with(
        [android_sdk.sdkmanager_path, "--list_installed"],
        env=android_sdk.env,
    )



import pytest


@pytest.mark.parametrize(
    "support_revision, host_arch, is_32bit_python, url",
    [
        (
            "3.10.9+20230116",
            "x86_64",
            False,
            "20230116/cpython-3.10.9+20230116-x86_64-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.11.1+20230116",
            "aarch64",
            False,
            "20230116/cpython-3.11.1+20230116-aarch64-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.11.1+20230116",
            "aarch64",
            True,
            "20230116/cpython-3.11.1+20230116-armv7-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.11.1+20230116",
            "armv7l",
            True,
            "20230116/cpython-3.11.1+20230116-armv7l-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.8.16+20221220",
            "x86_64",
            False,
            "20221220/cpython-3.8.16+20221220-x86_64-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.8.16+20221220",
            "x86_64",
            True,
            "20221220/cpython-3.8.16+20221220-i686-unknown-linux-gnu-install_only.tar.gz",
        ),
        (
            "3.8.16+20221220",
            "i686",
            True,
            "20221220/cpython-3.8.16+20221220-i686-unknown-linux-gnu-install_only.tar.gz",
        ),
    ],
)
def test_support_package_url(
    linux_mixin,
    host_arch,
    support_revision,
    is_32bit_python,
    url,
):
    """The support package URL is customized."""
    # Set up the host architecture
    linux_mixin.tools.host_arch = host_arch
    linux_mixin.tools.is_32bit_python = is_32bit_python

    assert linux_mixin.support_package_url(support_revision) == (
        "https://github.com/indygreg/python-build-standalone/releases/download/" + url
    )



import pytest

from briefcase.console import sanitize_text


@pytest.mark.parametrize(
    "input_text, sanitized_text",
    [
        (
            "log output",
            "log output",
        ),
        (
            "ls\n\x1b[00m\x1b[01;31mexamplefile.zip\x1b[00m\n\x1b[01;31m",
            "ls\nexamplefile.zip\n",
        ),
        (
            "log output: \u001b[31mRed\u001B[0m",
            "log output: Red",
        ),
        (
            "\u001b[1mbold log output:\u001b[0m \u001b[4mUnderline\u001b[0m",
            "bold log output: Underline",
        ),
        (
            f"{chr(7)}{chr(8)}{chr(11)}{chr(12)}{chr(13)}log{chr(7)} output{chr(7)}",
            "log output",
        ),
    ],
)
def test_sanitize_text(input_text, sanitized_text):
    """Text is sanitized as expected."""
    assert sanitize_text(input_text) == sanitized_text



import platform
from pathlib import Path

import pytest

# Default encoding is platform specific.
if platform.platform() == "Windows":
    CONSOLE_ENCODING = "cp437"
else:
    CONSOLE_ENCODING = "UTF-8"


def test_no_overrides(mock_sub):
    """With no overrides, there are still kwargs."""
    assert mock_sub.final_kwargs() == {
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
    }


def test_explicit_no_overrides(mock_sub):
    """With explicitly no overrides, there are still kwargs."""
    assert mock_sub.final_kwargs(env=None) == {
        "env": {
            "VAR1": "Value 1",
            "PS1": "\nLine 2\n\nLine 4",
            "PWD": "/home/user/",
        },
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
    }


def test_env_overrides(mock_sub):
    """If environmental overrides are provided, they supersede the default
    environment."""
    assert mock_sub.final_kwargs(env={"NEWVAR": "value", "VAR1": "New Value"}) == {
        "env": {
            "VAR1": "New Value",
            "PS1": "\nLine 2\n\nLine 4",
            "PWD": "/home/user/",
            "NEWVAR": "value",
        },
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
    }


def test_env_overrides_none(mock_sub):
    """Any overrides set to None are absent in the final kwargs."""
    assert mock_sub.final_kwargs(env={"NEWVAR": "val", "VAR1": None, "VAR2": None}) == {
        "env": {
            "PS1": "\nLine 2\n\nLine 4",
            "PWD": "/home/user/",
            "NEWVAR": "val",
        },
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
    }


def test_cwd_provided(mock_sub):
    """If a cwd is provided, it is reflected in the environment."""
    cwd_override = "/my/current/path"
    assert mock_sub.final_kwargs(cwd=cwd_override) == {
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
        "cwd": cwd_override,
    }


def test_non_str_cwd_provided(mock_sub):
    """If the cwd isn't a string, it's converted to string."""
    cwd_override = Path("/my/current/path")
    assert mock_sub.final_kwargs(cwd=cwd_override) == {
        "text": True,
        "encoding": CONSOLE_ENCODING,
        "errors": "backslashreplace",
        "cwd": str(cwd_override),
    }


@pytest.mark.parametrize(
    "in_kwargs, final_kwargs",
    [
        # Default handling
        (
            {},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "backslashreplace"},
        ),
        # Explicit text/universal_newlines provided
        (
            {"text": True},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "backslashreplace"},
        ),
        ({"text": False}, {"text": False}),
        ({"universal_newlines": False}, {"text": False}),
        (
            {"universal_newlines": True},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "backslashreplace"},
        ),
        # Explicit encoding provided
        (
            {"encoding": "ibm850"},
            {"text": True, "encoding": "ibm850", "errors": "backslashreplace"},
        ),
        (
            {"text": True, "encoding": "ibm850"},
            {"text": True, "encoding": "ibm850", "errors": "backslashreplace"},
        ),
        (
            {"text": False, "encoding": "ibm850"},
            {"text": True, "encoding": "ibm850", "errors": "backslashreplace"},
        ),
        (
            {"universal_newlines": False, "encoding": "ibm850"},
            {"text": True, "encoding": "ibm850", "errors": "backslashreplace"},
        ),
        (
            {"universal_newlines": True, "encoding": "ibm850"},
            {"text": True, "encoding": "ibm850", "errors": "backslashreplace"},
        ),
        # Explicit errors provided
        (
            {"errors": "emojireplace"},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "emojireplace"},
        ),
        (
            {"encoding": "ascii", "errors": "emojireplace"},
            {"text": True, "encoding": "ascii", "errors": "emojireplace"},
        ),
        (
            {"text": True, "errors": "emojireplace"},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "emojireplace"},
        ),
        (
            {"text": False, "errors": "emojireplace"},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "emojireplace"},
        ),
        (
            {"universal_newlines": True, "errors": "emojireplace"},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "emojireplace"},
        ),
        (
            {"universal_newlines": False, "errors": "emojireplace"},
            {"text": True, "encoding": CONSOLE_ENCODING, "errors": "emojireplace"},
        ),
    ],
)
def test_text_conversion(mock_sub, in_kwargs, final_kwargs):
    """Text/universal_newlines is correctly inserted/overridden, with encoding."""
    assert mock_sub.final_kwargs(**in_kwargs) == final_kwargs



import sys
from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.platforms.android.gradle import GradleCreateCommand


@pytest.fixture
def create_command(tmp_path, first_app_config):
    command = GradleCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    return command


@pytest.mark.parametrize("host_os", ["WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(UnsupportedHostError, match="This command is not supported on"):
        create_command()


def test_unsupported_template_version(create_command, first_app_config):
    """Error raised if template's target version is not supported."""
    # Skip rolling out the template and support package
    create_command.generate_app_template = MagicMock()
    create_command.install_app_support_package = MagicMock()

    # Skip tool verification
    create_command.verify_tools = MagicMock()

    create_command.verify_app = MagicMock(wraps=create_command.verify_app)

    create_command._briefcase_toml.update(
        {first_app_config: {"briefcase": {"target_epoch": "0.3.16"}}}
    )

    with pytest.raises(
        BriefcaseCommandError,
        match="The app template used to generate this app is not compatible",
    ):
        create_command(first_app_config)

    create_command.verify_app.assert_called_once_with(first_app_config)


def test_support_package_filename(create_command):
    """The Android support package filename has been customized."""
    assert (
        create_command.support_package_filename(52)
        == f"Python-3.{sys.version_info.minor}-Android-support.b52.zip"
    )


@pytest.mark.parametrize(
    "version, build, version_code",
    [
        ("0.1", None, "10000"),
        ("0.1a3", None, "10000"),
        ("1.2", None, "1020000"),
        ("1.2a3", None, "1020000"),
        ("1.2.3", None, "1020300"),
        ("1.2.3a3", None, "1020300"),
        ("1.2.3b4", None, "1020300"),
        ("1.2.3rc5", None, "1020300"),
        ("1.2.3.dev6", None, "1020300"),
        ("1.2.3.post7", None, "1020300"),
        # Date based
        ("2019.1", None, "2019010000"),
        ("2019.18", None, "2019180000"),
        ("2019.4.18", None, "2019041800"),
        # Build number can be injected
        ("0.1", 3, "10003"),
        ("0.1a3", 42, "10042"),
        ("1.2", 42, "1020042"),
        ("1.2a3", 3, "1020003"),
        ("1.2.3", 3, "1020303"),
        ("1.2.3b4", 42, "1020342"),
        ("2019.1", 3, "2019010003"),
        ("2019.1b4", 42, "2019010042"),
    ],
)
def test_version_code(create_command, first_app_config, version, build, version_code):
    """Validate that create adds version_code to the template context."""
    first_app_config.version = version
    if build:
        first_app_config.build = build
    context = create_command.output_format_template_context(first_app_config)
    assert context["version_code"] == version_code
    assert context["safe_formal_name"] == "First App"

    # Version code must be less than a 32-bit signed integer MAXINT.
    assert int(version_code) < 2147483647


@pytest.mark.parametrize(
    "input, output, has_warning",
    [
        (
            None,
            {
                "implementation": [
                    "androidx.appcompat:appcompat:1.0.2",
                    "androidx.constraintlayout:constraintlayout:1.1.3",
                    "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
                ]
            },
            True,
        ),
        (
            [],
            {"implementation": []},
            False,
        ),
        (
            [
                "com.example.foo:foo:1.2.3",
                "com.example.bar:bar:2.3.4",
            ],
            {
                "implementation": [
                    "com.example.foo:foo:1.2.3",
                    "com.example.bar:bar:2.3.4",
                ]
            },
            False,
        ),
    ],
)
def test_build_gradle_dependencies(
    create_command,
    first_app_config,
    input,
    output,
    has_warning,
    capsys,
):
    """Validate that create adds version_code to the template context."""
    if input is not None:
        first_app_config.build_gradle_dependencies = input

    context = create_command.output_format_template_context(first_app_config)
    assert context["build_gradle_dependencies"] == output

    assert (
        "** WARNING: App does not define build_gradle_dependencies              **"
        in capsys.readouterr().out
    ) == has_warning


extract_packages_params = [
    ([], ""),
    ([""], ""),
    (["one"], '"one"'),
    (["one/two"], '"two"'),
    (["one//two"], '"two"'),
    (["one/two/three"], '"three"'),
    (["one", "two"], '"one", "two"'),
    (["one", "two", "three"], '"one", "two", "three"'),
    (["one/two", "three/four"], '"two", "four"'),
    (["/leading"], '"leading"'),
    (["/leading/two"], '"two"'),
    (["/leading/two/three"], '"three"'),
    (["trailing/"], '"trailing"'),
    (["trailing//"], '"trailing"'),
    (["trailing/two/"], '"two"'),
]

# Handle differences in UNC path parsing (https://github.com/python/cpython/pull/100351).
extract_packages_params += [
    (
        ["//leading"],
        "" if sys.platform == "win32" and sys.version_info >= (3, 12) else '"leading"',
    ),
    (
        ["//leading/two"],
        "" if sys.platform == "win32" else '"two"',
    ),
    (["//leading/two/three"], '"three"'),
    (["//leading/two/three/four"], '"four"'),
]

if sys.platform == "win32":
    extract_packages_params += [
        ([path.replace("/", "\\") for path in test_sources], expected)
        for test_sources, expected in extract_packages_params
    ]


@pytest.mark.parametrize("test_sources, expected", extract_packages_params)
def test_extract_packages(create_command, first_app_config, test_sources, expected):
    first_app_config.test_sources = test_sources
    context = create_command.output_format_template_context(first_app_config)
    assert context["extract_packages"] == expected


@pytest.mark.parametrize(
    "permissions, features, context",
    [
        # No permissions
        (
            {},
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                },
                "features": {},
            },
        ),
        # Only custom permissions
        (
            {
                "android.permission.READ_CONTACTS": True,
            },
            {
                "android.hardware.bluetooth": True,
            },
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.READ_CONTACTS": True,
                },
                "features": {
                    "android.hardware.bluetooth": True,
                },
            },
        ),
        # Camera permissions
        (
            {
                "camera": "I need to see you",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.CAMERA": True,
                },
                "features": {
                    "android.hardware.camera": False,
                    "android.hardware.camera.any": False,
                    "android.hardware.camera.autofocus": False,
                    "android.hardware.camera.external": False,
                    "android.hardware.camera.front": False,
                },
            },
        ),
        # Microphone permissions
        (
            {
                "microphone": "I need to hear you",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.RECORD_AUDIO": True,
                },
                "features": {},
            },
        ),
        # Coarse location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_COARSE_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Fine location permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_FINE_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Background location permissions
        (
            {
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_BACKGROUND_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Coarse location background permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_COARSE_LOCATION": True,
                    "android.permission.ACCESS_BACKGROUND_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Fine location background permissions
        (
            {
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_FINE_LOCATION": True,
                    "android.permission.ACCESS_BACKGROUND_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Coarse and fine location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_COARSE_LOCATION": True,
                    "android.permission.ACCESS_FINE_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Coarse and fine background location permissions
        (
            {
                "coarse_location": "I need to know roughly where you are",
                "fine_location": "I need to know exactly where you are",
                "background_location": "I always need to know where you are",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.ACCESS_COARSE_LOCATION": True,
                    "android.permission.ACCESS_FINE_LOCATION": True,
                    "android.permission.ACCESS_BACKGROUND_LOCATION": True,
                },
                "features": {
                    "android.hardware.location.gps": False,
                    "android.hardware.location.network": False,
                },
            },
        ),
        # Photo library permissions
        (
            {
                "photo_library": "I need to see your library",
            },
            {},
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.READ_MEDIA_VISUAL_USER_SELECTED": True,
                },
                "features": {},
            },
        ),
        # Override and augment by cross-platform definitions
        (
            {
                "camera": "I need to see you",
                "android.permission.CAMERA": False,
                "android.permission.READ_CONTACTS": True,
            },
            {
                "android.hardware.camera.external": True,
                "android.hardware.bluetooth": True,
            },
            {
                "permissions": {
                    "android.permission.ACCESS_NETWORK_STATE": True,
                    "android.permission.INTERNET": True,
                    "android.permission.CAMERA": False,
                    "android.permission.READ_CONTACTS": True,
                },
                "features": {
                    "android.hardware.camera": False,
                    "android.hardware.camera.any": False,
                    "android.hardware.camera.autofocus": False,
                    "android.hardware.camera.external": True,
                    "android.hardware.camera.front": False,
                    "android.hardware.bluetooth": True,
                },
            },
        ),
    ],
)
def test_permissions_context(create_command, first_app, permissions, features, context):
    """Platform-specific permissions can be added to the context."""
    # Set the permission and entitlement value
    first_app.permission = permissions
    first_app.feature = features
    # Extract the cross-platform permissions
    x_permissions = create_command._x_permissions(first_app)
    # Check that the final platform permissions are rendered as expected.
    assert context == create_command.permissions_context(first_app, x_permissions)



from unittest.mock import PropertyMock

import pytest
from rich.console import ColorSystem

from briefcase.console import Console, Printer


@pytest.mark.parametrize(
    "no_color, color_system, is_enabled",
    [
        (False, ColorSystem.TRUECOLOR, True),
        (False, None, False),
        (True, ColorSystem.TRUECOLOR, False),
        (True, None, False),
    ],
)
def test_is_color_enabled(no_color, color_system, is_enabled, monkeypatch):
    """Color is enabled/disabled based on no_color and color_system."""
    printer = Printer()
    monkeypatch.setattr(
        type(printer.console),
        "color_system",
        PropertyMock(return_value=color_system),
    )
    printer.console.no_color = no_color
    console = Console(printer=printer)

    # confirm these values to make sure they didn't change somehow...
    assert printer.console.no_color is no_color
    assert printer.console.color_system is color_system

    assert console.is_color_enabled is is_enabled



import os
import platform
import subprocess
from datetime import date
from unittest import mock

import pytest
from cookiecutter import exceptions as cookiecutter_exceptions
from git import exc as git_exceptions

import briefcase
from briefcase.exceptions import (
    BriefcaseCommandError,
    NetworkFailure,
    TemplateUnsupportedVersion,
)


@pytest.fixture
def full_context():
    return {
        "format": "dummy",
        "app_name": "my-app",
        "formal_name": "My App",
        "bundle": "com.example",
        "bundle_identifier": "com.example.my-app",
        "version": "1.2.3",
        "description": "This is a simple app",
        "long_description": None,
        "sources": ["src/my_app"],
        "test_sources": None,
        "test_requires": None,
        "url": "https://example.com",
        "author": "First Last",
        "author_email": "first@example.com",
        "requires": None,
        "icon": None,
        "supported": True,
        "console_app": False,
        "permissions": {},
        "custom_permissions": {},
        "requests": {},
        "document_types": {},
        "license": {"file": "LICENSE"},
        # Properties of the generating environment
        "python_version": platform.python_version(),
        "host_arch": "gothic",
        "briefcase_version": briefcase.__version__,
        # Properties of the template
        "template_source": "https://github.com/beeware/briefcase-Tester-Dummy-template.git",
        "template_branch": f"v{briefcase.__version__}",
        # Fields generated from other properties
        "module_name": "my_app",
        "class_name": "MyApp",
        "package_name": "com.example",
        # Date-based fields added at time of generation
        "year": date.today().strftime("%Y"),
        "month": date.today().strftime("%B"),
        # Fields added by the output format.
        "output_format": "dummy",
        # These tests don't do a full finalization, so the context will still be
        # marked as draft.
        "__draft__": True,
    }


@pytest.mark.parametrize(
    "briefcase_version, expected_branch",
    [
        ("37.42.1", "v37.42.1"),
        ("37.42.2.dev0+gad61a29.d20220919", "v37.42.2"),
        ("37.42.3.dev73+gad61a29.d20220919", "v37.42.3"),
        ("37.42.4a1", "v37.42.4"),
        ("37.42.5b2", "v37.42.5"),
        ("37.42.6rc3", "v37.42.6"),
        ("37.42.7.post1", "v37.42.7"),
    ],
)
def test_default_template(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    briefcase_version,
    expected_branch,
    tmp_path,
):
    """Absent of other information, the briefcase version (without suffixes) is used as
    the template branch."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    full_context["briefcase_version"] = briefcase_version
    full_context["template_branch"] = expected_branch

    # There won't be a cookiecutter cache, so there won't be a cache path (yet).

    # Generate the template.
    create_command.generate_app_template(myapp)
    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout=expected_branch,
        output_dir=str(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


@pytest.mark.parametrize(
    "briefcase_version,template_branch",
    [
        ("37.42.7.dev0+gad61a29.d20220919", "v37.42.7"),
        ("37.42.7.dev73+gad61a29.d20220919", "v37.42.7"),
    ],
)
def test_default_template_dev(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    briefcase_version,
    template_branch,
    tmp_path,
):
    """In a dev version, template will fall back to the 'main' branch if a versioned
    template doesn't exist."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    full_context["briefcase_version"] = briefcase_version
    full_context["template_branch"] = template_branch

    full_context2 = full_context.copy()
    full_context2["template_branch"] = "main"

    # There won't be a cookiecutter cache, so there won't be a cache path (yet).

    # There will be two calls to cookiecutter; one on the versioned branch,
    # and one on the `main` branch. The first call will fail because the
    # template doesn't exist yet; the second will succeed.
    create_command.tools.cookiecutter.side_effect = [
        cookiecutter_exceptions.RepositoryCloneFailed,
        None,
    ]

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    assert create_command.tools.cookiecutter.mock_calls == [
        mock.call(
            str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
            no_input=True,
            checkout="v37.42.7",
            output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
            extra_context=full_context,
            default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
        ),
        mock.call(
            str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
            no_input=True,
            checkout="main",
            output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
            extra_context=full_context2,
            default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
        ),
    ]


@pytest.mark.parametrize(
    "briefcase_version",
    ("37.42.7.dev0+gad61a29.d20220919", "37.42.7.dev73+gad61a29.d20220919"),
)
def test_default_template_dev_explicit_branch(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    briefcase_version,
    tmp_path,
):
    """In a dev version, if an explicit branch is provided, it is used."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    full_context["briefcase_version"] = briefcase_version

    # Set an explicit branch
    branch = "some_branch"
    myapp.template_branch = branch
    full_context["template_branch"] = branch

    # There won't be a cookiecutter cache, so there won't be
    # a cache path (yet).
    create_command.tools.git.Repo.side_effect = git_exceptions.NoSuchPathError

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked (once) with the expected template name and context.
    assert create_command.tools.cookiecutter.mock_calls == [
        mock.call(
            str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
            no_input=True,
            checkout=branch,
            output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
            extra_context=full_context,
            default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
        ),
    ]


@pytest.mark.parametrize(
    "briefcase_version",
    ("37.42.7.dev0+gad61a29.d20220919", "37.42.7.dev73+gad61a29.d20220919"),
)
def test_default_template_dev_explicit_invalid_branch(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    briefcase_version,
    tmp_path,
):
    """In a dev version, if an explicit (but invalid) branch is provided, the fallback
    to the 'main' branch will not occur."""
    # Set the Briefcase version to a dev version
    monkeypatch.setattr(briefcase, "__version__", briefcase_version)
    full_context["briefcase_version"] = briefcase_version

    # Set an explicit branch
    branch = "some_branch"
    myapp.template_branch = branch
    full_context["template_branch"] = branch

    # There won't be a cookiecutter cache, so there won't be
    # a cache path (yet).
    create_command.tools.git.Repo.side_effect = git_exceptions.NoSuchPathError

    # There will be only 1 call to cookiecutter; the one on the versioned branch.
    create_command.tools.cookiecutter.side_effect = (
        cookiecutter_exceptions.RepositoryCloneFailed
    )

    # Generate the template.
    with pytest.raises(TemplateUnsupportedVersion):
        create_command.generate_app_template(myapp)

    # Cookiecutter was invoked (once) with the expected template name and context.
    assert create_command.tools.cookiecutter.mock_calls == [
        mock.call(
            str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
            no_input=True,
            checkout=branch,
            output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
            extra_context=full_context,
            default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
        ),
    ]


def test_explicit_branch(monkeypatch, create_command, myapp, full_context, tmp_path):
    """User can choose which branch to take the template from."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"

    # Set an explicit branch
    branch = "some_branch"
    myapp.template_branch = branch
    full_context["template_branch"] = branch

    # There won't be a cookiecutter cache, so there won't be
    # a cache path (yet).
    create_command.tools.git.Repo.side_effect = git_exceptions.NoSuchPathError

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout=branch,
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_platform_exists(monkeypatch, create_command, myapp, full_context, tmp_path):
    """If the platform directory already exists, it's ok."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    # There won't be a cookiecutter cache, so there won't be
    # a cache path (yet).
    create_command.tools.git.Repo.side_effect = git_exceptions.NoSuchPathError

    # Create the build directory
    create_command.build_path(myapp).mkdir(parents=True)

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_explicit_repo_template(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If a template is specified in the app config, it is used."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    template = "https://example.com/magic/special-template.git"
    myapp.template = template
    full_context["template_source"] = template

    # There won't be a cookiecutter cache, so there won't be a repo path (yet).

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/special-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_explicit_repo_template_and_branch(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If a template and branch is specified in the app config, it is used."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"

    # Set an explicit template and branch
    template = "https://example.com/magic/special-template.git"
    myapp.template = template
    full_context["template_source"] = template

    branch = "some_branch"
    myapp.template_branch = branch
    full_context["template_branch"] = branch

    # There won't be a cookiecutter cache, so there won't be a repo path (yet).

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/special-template"),
        no_input=True,
        checkout=branch,
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_explicit_local_template(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If a local template path is specified in the app config, it is used."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    template = "/path/to/special-template"
    myapp.template = template
    full_context["template_source"] = template

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        "/path/to/special-template",
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )

    # The template is a local directory, so there won't be any calls on git.
    assert create_command.tools.git.Repo.call_count == 0


def test_explicit_local_template_and_branch(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If a local template path and branch is specified in the app config, it is
    used."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"

    template = "/path/to/special-template"
    myapp.template = template
    full_context["template_source"] = template

    branch = "some_branch"
    myapp.template_branch = branch
    full_context["template_branch"] = branch

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        "/path/to/special-template",
        no_input=True,
        checkout=branch,
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )

    # The template is a local directory, so there won't be any calls on git.
    assert create_command.tools.git.Repo.call_count == 0


def test_offline_repo_template(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If the user is offline the first time a repo template is requested, an error is
    raised."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    # There won't be a cookiecutter cache, so there won't be a repo path (yet).

    # Calling cookiecutter on a repository while offline causes a CalledProcessError
    create_command.tools.cookiecutter.side_effect = subprocess.CalledProcessError(
        cmd=[
            "git",
            "clone",
            str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        ],
        returncode=128,
    )

    # Generating the template under these conditions raises an error
    with pytest.raises(NetworkFailure):
        create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_invalid_repo_template(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    """If the provided template URL isn't valid, an error is raised."""
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    template = "https://example.com/somewhere/not-a-repo.git"
    myapp.template = template
    full_context["template_source"] = template

    # Prime git to respond to a bad URL.
    create_command.tools.git.Repo.clone_from.side_effect = (
        git_exceptions.GitCommandError("git", 128)
    )

    # Generating the template under there conditions raises an error
    with pytest.raises(BriefcaseCommandError, match=r"..."):
        create_command.generate_app_template(myapp)

    # No attempt is made to create a template.
    create_command.tools.cookiecutter.assert_not_called()


def test_cached_template(monkeypatch, create_command, myapp, full_context, tmp_path):
    """If a template has already been used, the cached version will be used."""
    # Set up a pre-cached template directory
    (create_command.data_path / "templates/briefcase-Tester-Dummy-template").mkdir(
        parents=True
    )

    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()
    mock_remote_head = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has
    # heads that can be accessed.
    create_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.refs.__getitem__.return_value = mock_remote_head

    # Generate the template.
    create_command.generate_app_template(myapp)

    # The origin of the repo was fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.fetch.assert_called_once_with()

    # The remote head was checked out.
    mock_remote_head.checkout.assert_called_once_with()

    # Cookiecutter was invoked with the path to the *cached* template name
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_cached_template_offline(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    capsys,
    tmp_path,
):
    """If the user is offline, a cached template won't be updated, but will still
    work."""
    # Set up a pre-cached template directory
    (create_command.data_path / "templates/briefcase-Tester-Dummy-template").mkdir(
        parents=True
    )

    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()
    mock_remote_head = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has
    # heads that can be accessed. However, calling fetch on the remote
    # will cause a git error (error code 128).
    create_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.refs.__getitem__.return_value = mock_remote_head
    mock_remote.fetch.side_effect = git_exceptions.GitCommandError("git", 128)

    # Generate the template.
    create_command.generate_app_template(myapp)

    # An attempt to fetch the repo origin was made
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.fetch.assert_called_once_with()

    # A warning was raised to the user about the fetch problem
    output = capsys.readouterr().out
    assert "** WARNING: Unable to update template" in output

    # The remote head was checked out.
    mock_remote_head.checkout.assert_called_once_with()

    # Cookiecutter was invoked with the path to the *cached* template name
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=str(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )


def test_cached_missing_branch_template(monkeypatch, create_command, myapp):
    """If the cached repo doesn't have a branch for this Briefcase version, an error is
    raised."""
    # Set up a pre-cached template directory
    (create_command.data_path / "templates/briefcase-Tester-Dummy-template").mkdir(
        parents=True
    )

    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote. However, it doesn't
    # have a head corresponding to the requested Python version, so it
    # raises an IndexError
    create_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.refs.__getitem__.side_effect = IndexError

    # Generating the template under there conditions raises an error
    with pytest.raises(TemplateUnsupportedVersion):
        create_command.generate_app_template(myapp)


def test_x_permissions(
    monkeypatch,
    create_command,
    myapp,
    full_context,
    tmp_path,
):
    # Set the Briefcase version
    monkeypatch.setattr(briefcase, "__version__", "37.42.7")
    full_context["briefcase_version"] = "37.42.7"
    full_context["template_branch"] = "v37.42.7"

    # Define some permissions and requests. The original "permission" and "request"
    # definitions will be hidden from the final template context.

    myapp.permission = {
        # Cross-platform permissions
        "camera": "I need to see you",
        "microphone": "I need to hear you",
        "coarse_location": "I need to know approximately where you are",
        "fine_location": "I need to know exactly where you are",
        "background_location": "I need to know where you are constantly",
        "photo_library": "I need to see your photos",
        # Custom permissions
        "DUMMY_sit": "I can't sit without an invitation",
        "DUMMY.leave.the.dinner.table": "It would be impolite.",
    }
    myapp.request = {"tasty.beverage": True}

    # In the final context, all cross-platform permissions have been converted to upper
    # case, prefixed with "DUMMY", and moved to the `permissions` key. Custom
    # permissions have been moved to the "custom_permissions" key
    full_context["permissions"] = {
        "DUMMY_CAMERA": "I NEED TO SEE YOU",
        "DUMMY_MICROPHONE": "I NEED TO HEAR YOU",
        "DUMMY_COARSE_LOCATION": "I NEED TO KNOW APPROXIMATELY WHERE YOU ARE",
        "DUMMY_FINE_LOCATION": "I NEED TO KNOW EXACTLY WHERE YOU ARE",
        "DUMMY_BACKGROUND_LOCATION": "I NEED TO KNOW WHERE YOU ARE CONSTANTLY",
        "DUMMY_PHOTO_LIBRARY": "I NEED TO SEE YOUR PHOTOS",
    }
    full_context["custom_permissions"] = {
        "DUMMY_sit": "I can't sit without an invitation",
        "DUMMY.leave.the.dinner.table": "It would be impolite.",
    }

    # An extra request has been added because of the camera permission, and the
    # custom request has been preserved.
    full_context["requests"] = {
        "good.lighting": True,
        "tasty.beverage": True,
    }

    # There won't be a cookiecutter cache, so there won't be
    # a cache path (yet).
    create_command.tools.git.Repo.side_effect = git_exceptions.NoSuchPathError

    # Generate the template.
    create_command.generate_app_template(myapp)

    # Cookiecutter was invoked with the expected template name and context.
    create_command.tools.cookiecutter.assert_called_once_with(
        str(create_command.data_path / "templates/briefcase-Tester-Dummy-template"),
        no_input=True,
        checkout="v37.42.7",
        output_dir=os.fsdecode(tmp_path / "base_path/build/my-app/tester"),
        extra_context=full_context,
        default_config={"replay_dir": str(tmp_path / "data/templates/.replay")},
    )



def test_wait_bar_always_interactive(console):
    """Progress Bar is not disabled when console is interactive."""
    with console.progress_bar() as bar:
        assert bar.disable is False


def test_wait_bar_non_interactive(non_interactive_console):
    """Progress Bar is disabled when console is non-interactive."""
    with non_interactive_console.progress_bar() as bar:
        assert bar.disable is True



import shutil
from unittest import mock

import pytest
from git import exc as git_exceptions

from briefcase.exceptions import BriefcaseCommandError, TemplateUnsupportedVersion

from ...utils import create_file


def test_non_url(base_command, mock_git):
    """If the provided template isn't a URL, don't try to update."""
    base_command.tools.git = mock_git

    cached_template = base_command.update_cookiecutter_cache(
        template="/path/to/template",
        branch="special",
    )

    assert cached_template == "/path/to/template"
    # No git actions are performed
    assert base_command.tools.git.Repo.call_count == 0


def test_new_repo_template(base_command, mock_git):
    """If a previously unknown URL template is specified it is used."""
    base_command.tools.git = mock_git

    # There won't be a cookiecutter cache, so there won't be
    # a repo path (yet).

    cached_template = base_command.update_cookiecutter_cache(
        template="https://example.com/magic/special-template.git",
        branch="special",
    )

    # The template that will be used is the original URL
    assert cached_template == base_command.data_path / "templates" / "special-template"

    # A shallow clone is performed.
    base_command.tools.git.Repo.clone_from.assert_called_once_with(
        url="https://example.com/magic/special-template.git",
        to_path=base_command.data_path / "templates" / "special-template",
        filter=["blob:none"],
        no_checkout=True,
    )


def test_new_repo_template_interrupt(base_command, mock_git):
    """If the user raises a keyboard interrupt while cloning, the template is cleaned
    up."""
    base_command.tools.git = mock_git

    # Raise a KeyboardInterrupt during a the clone, having written the git config file.
    def clone_failure(to_path, **kwargs):
        create_file(to_path / ".git" / "config", "git config")
        raise KeyboardInterrupt()

    # Prime the error when the clone is interrupted
    base_command.tools.git.Repo.clone_from.side_effect = clone_failure

    with pytest.raises(KeyboardInterrupt):
        base_command.update_cookiecutter_cache(
            template="https://example.com/magic/special-template.git",
            branch="special",
        )

    # The template directory should be cleaned up
    assert not (base_command.data_path / "templates" / "special-template").exists()


def test_new_repo_template_mkdir_interrupt(base_command, mock_git):
    """A really early interrupt will occur before the template dir is created."""
    base_command.tools.git = mock_git

    # We don't have a convenient point to insert a KeyboardInterrupt *before* creating the
    # directory, so we fake the effect - make the side effect of the clone deletion of
    # the entire folder; then raise the KeyboardInterrupt.
    def clone_failure(to_path, **kwargs):
        shutil.rmtree(to_path)
        raise KeyboardInterrupt()

    # Prime the error when the clone is interrupted
    base_command.tools.git.Repo.clone_from.side_effect = clone_failure

    with pytest.raises(KeyboardInterrupt):
        base_command.update_cookiecutter_cache(
            template="https://example.com/magic/special-template.git",
            branch="special",
        )

    # The template directory shouldn't exist
    assert not (base_command.data_path / "templates" / "special-template").exists()


def test_new_repo_invalid_template_url(base_command, mock_git):
    """If a previously unknown URL template is specified it is used."""
    base_command.tools.git = mock_git

    # Prime the error when the repo doesn't exist
    base_command.tools.git.Repo.clone_from.side_effect = git_exceptions.GitCommandError(
        "git", 128
    )

    with pytest.raises(
        BriefcaseCommandError,
        match=(
            r"Unable to clone repository 'https://example.com/magic/special-template.git'"
        ),
    ):
        base_command.update_cookiecutter_cache(
            template="https://example.com/magic/special-template.git",
            branch="special",
        )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.clone_from.assert_called_once_with(
        url="https://example.com/magic/special-template.git",
        to_path=base_command.data_path / "templates" / "special-template",
        filter=["blob:none"],
        no_checkout=True,
    )

    # The template directory should be cleaned up
    assert not (base_command.data_path / "templates" / "special-template").exists()


def test_existing_repo_template(base_command, mock_git):
    """If a previously known URL template is specified it is used."""
    base_command.tools.git = mock_git

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()
    mock_remote_head = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has
    # heads that can be accessed.
    base_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.refs.__getitem__.return_value = mock_remote_head
    mock_remote.url = "https://example.com/magic/special-template.git"

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )
    cached_path.mkdir(parents=True)

    # Update the cache
    cached_template = base_command.update_cookiecutter_cache(
        template="https://example.com/magic/special-template.git",
        branch="special",
    )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # The origin of the repo was updated and fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.set_url.assert_called_once_with(
        new_url="https://example.com/magic/special-template.git",
        old_url="https://example.com/magic/special-template.git",
    )
    mock_remote.fetch.assert_called_once_with()

    # The right branch was accessed
    mock_remote.refs.__getitem__.assert_called_once_with("special")

    # The remote head was checked out.
    mock_remote_head.checkout.assert_called_once_with()

    # The template that will be used is the original URL
    assert cached_template == cached_path


def test_existing_repo_template_corrupted(base_command, mock_git):
    """If a previously cached URL template is in a corrupted state, it is deleted and
    re-cloned."""
    # Git returns an exception wrapping the given URL.
    base_command.tools.git = mock_git
    base_command.tools.git.Repo.side_effect = [
        git_exceptions.GitCommandError("git", 128),
        None,
    ]

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )

    # Create a bad cached template
    create_file(cached_path / "bad-template", "Bad template")

    cached_template = base_command.update_cookiecutter_cache(
        template="https://example.com/magic/special-template.git",
        branch="special",
    )

    # The template that will be used is the original URL
    assert cached_template == base_command.data_path / "templates" / "special-template"

    # An attempt was made to wrap the old repo
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # A shallow clone is performed.
    base_command.tools.git.Repo.clone_from.assert_called_once_with(
        url="https://example.com/magic/special-template.git",
        to_path=base_command.data_path / "templates" / "special-template",
        filter=["blob:none"],
        no_checkout=True,
    )

    # The old template content has been deleted
    assert not (cached_path / "bad-template").exists()


def test_existing_repo_template_with_different_url(base_command, mock_git):
    """If a previously known URL template is specified but uses a different remote URL,
    the repo's origin URL is updated and is used."""
    base_command.tools.git = mock_git

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()
    mock_remote_head = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has
    # heads that can be accessed.
    base_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.refs.__getitem__.return_value = mock_remote_head
    mock_remote.url = "https://example.com/existing/special-template.git"

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )
    cached_path.mkdir(parents=True)

    # Update the cache
    cached_template = base_command.update_cookiecutter_cache(
        template="https://example.com/magic/special-template.git",
        branch="special",
    )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # The origin of the repo was updated and fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.set_url.assert_called_once_with(
        new_url="https://example.com/magic/special-template.git",
        old_url="https://example.com/existing/special-template.git",
    )
    mock_remote.fetch.assert_called_once_with()

    # The right branch was accessed
    mock_remote.refs.__getitem__.assert_called_once_with("special")

    # The remote head was checked out.
    mock_remote_head.checkout.assert_called_once_with()

    # The template that will be used is the original URL
    assert cached_template == cached_path


def test_offline_repo_template(base_command, mock_git, capsys):
    """If the user is offline the when a repo template is requested, but the branch
    exists, the command continues with a warning."""
    base_command.tools.git = mock_git

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()
    mock_remote_head = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has
    # heads that can be accessed. However, calling fetch on the remote
    # will cause a git error (error code 128).
    base_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.url = "https://example.com/magic/special-template.git"
    mock_remote.refs.__getitem__.return_value = mock_remote_head
    mock_remote.fetch.side_effect = git_exceptions.GitCommandError("git", 128)

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )
    cached_path.mkdir(parents=True)

    # Update the cache
    cached_template = base_command.update_cookiecutter_cache(
        template="https://example.com/magic/special-template.git", branch="special"
    )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # The origin of the repo was updated and fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.set_url.assert_called_once_with(
        new_url="https://example.com/magic/special-template.git",
        old_url="https://example.com/magic/special-template.git",
    )
    mock_remote.fetch.assert_called_once_with()

    # The right branch was accessed
    mock_remote.refs.__getitem__.assert_called_once_with("special")

    # The remote head was checked out.
    mock_remote_head.checkout.assert_called_once_with()

    # The template that will be used is the original URL
    assert cached_template == cached_path

    # A warning was raised about the template possibly being stale.
    assert "WARNING: Unable to update template" in capsys.readouterr().out


def test_cached_missing_branch_template(base_command, mock_git):
    """If the cached repo doesn't have the requested branch, an error is raised."""
    base_command.tools.git = mock_git

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote. However, it doesn't
    # have a head corresponding to the requested Python version, so it
    # raises an IndexError
    base_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.return_value = mock_remote
    mock_remote.url = "https://example.com/magic/special-template.git"
    mock_remote.refs.__getitem__.side_effect = IndexError

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )
    cached_path.mkdir(parents=True)

    # Generating the template under there conditions raises an error
    with pytest.raises(TemplateUnsupportedVersion):
        base_command.update_cookiecutter_cache(
            template="https://example.com/magic/special-template.git",
            branch="invalid",
        )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # The origin of the repo was updated and fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.set_url.assert_called_once_with(
        new_url="https://example.com/magic/special-template.git",
        old_url="https://example.com/magic/special-template.git",
    )
    mock_remote.fetch.assert_called_once_with()

    # An attempt to access the branch was made
    mock_remote.refs.__getitem__.assert_called_once_with("invalid")


def test_git_repo_with_missing_origin_remote(base_command, mock_git):
    """If the local git repo doesn't have an origin remote, a ValueError is raised."""
    base_command.tools.git = mock_git

    mock_repo = mock.MagicMock()
    mock_remote = mock.MagicMock()

    # Git returns a Repo, that repo can return a remote, and it has heads that can be
    # accessed. However, getting the remote will fail if git clone is not complete.
    base_command.tools.git.Repo.return_value = mock_repo
    mock_repo.remote.side_effect = ValueError("Remote named origin did not exist")

    cached_path = base_command.template_cache_path(
        "https://example.com/magic/special-template.git"
    )
    cached_path.mkdir(parents=True)

    # Update the cache
    with pytest.raises(
        BriefcaseCommandError,
        match="Unable to check out template branch.",
    ):
        base_command.update_cookiecutter_cache(
            template="https://example.com/magic/special-template.git", branch="special"
        )

    # The cookiecutter cache location will be interrogated.
    base_command.tools.git.Repo.assert_called_once_with(cached_path)

    # The origin of the repo was not updated or fetched
    mock_repo.remote.assert_called_once_with(name="origin")
    mock_remote.set_url.assert_not_called()
    mock_remote.fetch.assert_not_called()
